
# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT

from typing import Any
from typing import Dict
from typing import List
from typing import Tuple
from typing import Union

import json

from langchain_core.messages.ai import AIMessage
from langchain_core.messages.base import BaseMessage
from langchain_core.messages.human import HumanMessage
from langchain_core.messages.system import SystemMessage
from langchain_core.messages.tool import ToolMessage

from neuro_san.internals.messages.agent_message import AgentMessage
from neuro_san.internals.messages.agent_framework_message import AgentFrameworkMessage
from neuro_san.internals.messages.agent_tool_result_message import AgentToolResultMessage
from neuro_san.internals.messages.chat_message_type import ChatMessageType


def pretty_the_messages(messages: List[Any]) -> str:
    """
    Pretty printing helper
    :param messages: A list of OpenAI messages
    :return: A single string representing the output of the messages.
    """
    prettied = ""
    messages_list = list(messages)  # Convert to list to allow reversing

    # Iterate in reverse to find the last user message
    rm = []
    for m in reversed(messages_list):
        rm.append(m)
        if get_role(m) == "user":
            break

    for m in reversed(rm):
        if any(get_content(m)):
            prettied += f"{get_role(m)}: {get_content(m)}\n"

    return prettied


def get_last_message_with_content(messages: List[Any]) -> object:
    """
    Sometimes just indexing a message list by [-1]
    gives you a message that does not actually have any content.
    This method gives you the last messages that does have content.

    :param messages: input list of OpenAI messages
    :return: An applicalble OpenAI message (or None if no message is applicable)
    """

    messages_list = list(messages)

    for m in reversed(messages_list):
        if any(get_content(m)):
            return m

    return None


def generate_response(the_messages: List[Any]) -> str:
    """
    :param the_messages: A list of OpenAI messages
    :return: a JSON-ification of the list of messages.
    """
    response_list = []
    for i, m in enumerate(the_messages):
        # Duplicate the assistant role message before every tool response role message
        if get_role(m) == "tool" and i > 0 and get_role(the_messages[i - 1]) == "assistant":
            assistant_message = {
                "role": get_role(the_messages[i - 1]),
                "content": get_content(the_messages[i - 1])
            }
            response_list.append(assistant_message)

        message_dict = {
            "role": get_role(m),
            "content": get_content(m)
        }
        response_list.append(message_dict)

    return json.dumps(response_list)


def get_role(message: Any) -> str:
    """
    :param message: Either an OpenAI message or a langchain BaseMessage
    :return: A string describing the role of the message
    """

    if hasattr(message, "role"):
        return message.role

    # Check the look-up table above
    role: str = ChatMessageType.message_to_role(message)
    if role is not None:
        return role

    raise ValueError(f"Don't know how to handle message type {message.__class__.__name__}")


def get_content(message: Any) -> str:
    """
    :param message: Either an OpenAI message or a langchain BaseMessage
    :return: A string describing the content of the message
    """

    if isinstance(message, BaseMessage):
        # For OpenAI and Ollama, content of AI message is a string but content from
        # Anthropic AI message can either be a single string or a list of content blocks.
        # If it is a list, "text" is a key of a dictionary which is the first element of
        # the list. For more details: https://python.langchain.com/docs/integrations/chat/anthropic/#content-blocks
        content: Union[str, List] = message.content
        if isinstance(content, list):
            content = content[0].get("text", "")
        return content

    if hasattr(message, "content"):
        if not any(message.content):
            return ""
        return message.content[0].text.value

    raise ValueError(f"Don't know how to handle message type {message.__class__.__name__}")


def convert_to_chat_message(message: BaseMessage, origin: List[Dict[str, Any]] = None) -> Dict[str, Any]:
    """
    Convert the BaseMessage to a chat.ChatMessage dictionary

    :param message: The BaseMessage to convert
    :param origin: A List of origin dictionaries indicating the origin of the run.
            The origin can be considered a path to the original call to the front-man.
            Origin dictionaries themselves each have the following keys:
                "tool"                  The string name of the tool in the spec
                "instantiation_index"   An integer indicating which incarnation
                                        of the tool is being dealt with.
    :return: The ChatMessage in dictionary form
    """

    message_type: ChatMessageType = ChatMessageType.from_message(message)
    chat_message: Dict[str, Any] = {
        "type": message_type,
        # No mime_data for now
    }

    # Handle the origin information if we have it
    if origin is not None:
        chat_message["origin"] = origin

    # Dictionary of BaseMessage field sources to ChatMessage destinations
    # Anything in this dictionary is considered optional and we only populate
    # the field on ChatMessage if it has a value.
    optionals: Dict[str, str] = {
        "content": "text",
        "chat_context": "chat_context",
        "tool_result_origin": "tool_result_origin",
        "structure": "structure",
        "sly_data": "sly_data",
    }
    for src, dest in optionals.items():
        value: Any = None
        try:
            value = getattr(message, src)
        except AttributeError:
            # Not all BaseMessage subclasses have every field we are looking
            # for, and that is ok.
            value = None

        # For OpenAI and Ollama, content of AI message is a string but content from
        # Anthropic AI message can either be a single string or a list of content blocks.
        # If it is a list, "text" is a key of a dictionary which is the first element of
        # the list. For more details: https://python.langchain.com/docs/integrations/chat/anthropic/#content-blocks
        if src == "content" and isinstance(value, list):
            value = value[0].get("text", "")

        if value is not None:
            chat_message[dest] = value

    return chat_message


def convert_to_base_message(chat_message: Dict[str, Any], langchain_only: bool = True) -> BaseMessage:
    """
    The intended use for this method is for converting a neuro-san ChatMessage
    dictionary into a langchain BaseMessage such that it can be understood by
    a langchain agent within the context of a langchain-facing chat history.
    This means that some special messages we might have inserted into a chat
    history *might* need to be converted into something langchain agents will
    understand.

    :param chat_message: A ChatMessage dictionary to convert into BaseMessage
    :param langchain_only: When True (default) only convert langchain messages,
                        and not our own special variants.
    :return: A BaseMessage that was converted from the input.
            Can return None if conversion could not take place
    """
    base_message: BaseMessage = None
    if chat_message is None:
        return base_message

    content: str = chat_message.get("text")
    chat_message_type: ChatMessageType = ChatMessageType.from_response_type(chat_message.get("type"))

    if chat_message_type == ChatMessageType.SYSTEM:
        base_message = SystemMessage(content=content)
    elif chat_message_type == ChatMessageType.HUMAN:
        base_message = HumanMessage(content=content)
    elif chat_message_type == ChatMessageType.AI:
        base_message = AIMessage(content=content)
    elif chat_message_type == ChatMessageType.AGENT_TOOL_RESULT:
        base_message = AgentToolResultMessage(content=content,
                                              tool_result_origin=chat_message.get("tool_result_origin"))

    if langchain_only:
        # Go no further
        return base_message

    if base_message is None:
        # convert_to_base_message() is primarily for populating langchain chat history.
        # so handle these other types differently.
        if chat_message_type == ChatMessageType.AGENT:
            base_message = AgentMessage(content=chat_message.get("text", ""),
                                        structure=chat_message.get("structure"))
        elif chat_message_type == ChatMessageType.AGENT_FRAMEWORK:
            # Don't bother passing on chat_context
            base_message = AgentFrameworkMessage(content=chat_message.get("text", ""))

    # Any other message type we do not want to send to any agent as chat history.

    return base_message


def convert_to_message_tuple(base_message: BaseMessage) -> Tuple[str, Any]:
    """
    :param base_message: The base message to convert to a tuple used for a
                    prompt template.
    :return: A tuple corresponding to the input
    """
    if base_message is None:
        return None

    use_type: str = base_message.type
    if use_type == "agent_tool_result":
        # Langchain innards do not know about our own message types
        use_type = "ai"

    message_tuple: Tuple[str, Any] = (use_type, base_message.content)
    return message_tuple


def is_relevant_to_chat_history(base_message: BaseMessage) -> bool:
    """
    :param base_message: The BaseMessage to consider
    :return: True if the BaseMessage type is relevant to chat history (include).
             False otherwise (do not include).
    """
    if isinstance(base_message, (AgentMessage, AgentFrameworkMessage, ToolMessage)):
        # These guys cannot be in chat history as langchain will not recognize them.
        return False
    return True




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT

from typing import Any
from typing import Dict
from typing import List

from neuro_san.interfaces.concierge_session import ConciergeSession
from neuro_san.internals.tool_factories.service_tool_factory_provider import ServiceToolFactoryProvider


class DirectConciergeSession(ConciergeSession):
    """
    Service-agnostic guts for a ConciergeSession.
    This could be used by a gRPC and/or Http service.
    """

    # pylint: disable=too-many-arguments,too-many-positional-arguments
    def __init__(self,
                 metadata: Dict[str, Any] = None,
                 security_cfg: Dict[str, Any] = None):
        """
        Constructor

        :param metadata: A dictionary of request metadata to be forwarded
                        to subsequent yet-to-be-made requests.
        :param security_cfg: A dictionary of parameters used to
                        secure the TLS and the authentication of the gRPC
                        connection.  Supplying this implies use of a secure
                        GRPC Channel.  If None, uses insecure channel.
        """
        # These aren't used yet
        self._metadata: Dict[str, Any] = metadata
        self._security_cfg: Dict[str, Any] = security_cfg

    def list(self, request_dict: Dict[str, Any]) -> Dict[str, Any]:
        """
        :param request_dict: A dictionary version of the ConciergeRequest
                    protobuf structure. Has the following keys:
                        <None>
        :return: A dictionary version of the ConciergeResponse
                    protobuf structure. Has the following keys:
                "agents" - the sequence of dictionaries describing available agents
        """
        tool_factory_provider: ServiceToolFactoryProvider = \
            ServiceToolFactoryProvider.get_instance()
        agents_names: List[str] = tool_factory_provider.get_agent_names()
        agents_list: List[Dict[str, Any]] = []
        for agent_name in agents_names:
            agents_list.append({"agent_name": agent_name, "description": ""})
        return {"agents": agents_list}




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT
from typing import List
from typing import Set

from neuro_san.internals.errors.error_formatter_factory import ErrorFormatterFactory
from neuro_san.internals.interfaces.error_formatter import ErrorFormatter


class ErrorDetector:
    """
    Detects errors in agent output.
    """

    def __init__(self, agent_name: str,
                 error_formatter_name: str = None,
                 system_error_fragments: List[str] = None,
                 agent_error_fragments: List[str] = None):
        """
        Constructor

        :param agent_name: The name of the agent
        :param error_formatter_name: An optional name for which type of ErrorFormatter to use.
        :param system_error_fragments: A list of strings that identify errors from the system.
                By default this is None.
        :param agent_error_fragments: A list of strings that identify errors from the agent.
                By default this is None.
        """
        self.agent_name: str = agent_name
        self.error_formatter_name: str = error_formatter_name

        self.error_fragments: Set[str] = set()
        if system_error_fragments is not None:
            self.error_fragments.update(system_error_fragments)
        if agent_error_fragments is not None:
            self.error_fragments.update(agent_error_fragments)

    def is_error(self, output: str) -> bool:
        """
        Determines whether the output from the agent should be considered an error.

        :param output: String output from an agent
        :return: True/False as to whether or not this output should be considered an error.
        """
        if output is None:
            return False

        for fragment in self.error_fragments:
            if fragment in output:
                return True

        return False

    def handle_error(self, output: str, details: str = None) -> str:
        """
        Handles potential error in agent output.

        :param output: String output from an agent
        :param details: An optional details string to send to an ErrorFormatter if necessary.
        :return: The same string if there was no error, or if there was an error
                a properly formatted error per the choice in the error_formatter_name.
        """
        if not self.is_error(output):
            return output

        formatter: ErrorFormatter = ErrorFormatterFactory.create_formatter(self.error_formatter_name)
        error_output = formatter.format(self.agent_name, output, details)

        return error_output




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT
from typing import Any
from typing import List

from neuro_san.internals.graph.tools.calling_tool import CallingTool
from neuro_san.internals.interfaces.invocation_context import InvocationContext
from neuro_san.internals.run_context.interfaces.run import Run


class FrontMan(CallingTool):
    """
    A CallingTool implementation which is the root of the call graph.
    """

    async def submit_message(self, user_input: str) -> List[Any]:
        """
        Entry-point method for callers of the root of the Tool tree.

        :param user_input: An input string from the user.
        :return: A list of response messages for the run
        """
        # Initialize our return value
        decision_messages: List[Any] = []

        decision_run: Run = await self.run_context.submit_message(user_input)

        terminate = False
        while not terminate:
            if self.run_context is None:
                # Breaking from inside a container during cleanup can yield a None
                # run_context
                break

            decision_run = await self.run_context.wait_on_run(decision_run, self.journal)

            if decision_run.requires_action():
                decision_run = await self.make_tool_function_calls(decision_run)
            else:
                # Needs to get more information from the user on the basic task
                # of collecting information from the user about the decision.
                if self.run_context is None:
                    # Breaking from inside a container during cleanup can yield a None
                    # run_context
                    break
                decision_messages = await self.run_context.get_response()
                terminate = True

        return decision_messages

    def update_invocation_context(self, invocation_context: InvocationContext):
        """
        Update internal state based on the InvocationContext instance passed in.
        :param invocation_context: The context policy container that pertains to the invocation
        """
        self.journal = invocation_context.get_journal()
        if self.run_context is not None:
            self.run_context.update_invocation_context(invocation_context)

    async def build(self) -> List[Any]:
        """
        Main entry point to the class.

        :return: A List of messages produced during this process.
        """
        raise NotImplementedError




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT
from typing import Any
from typing import Awaitable
from typing import Dict
from typing import List
from typing import Union

from asyncio import Task
from contextvars import Context
from contextvars import ContextVar
from contextvars import copy_context
from time import time

from langchain_community.callbacks.openai_info import OpenAICallbackHandler
from langchain_community.callbacks.manager import get_openai_callback
from langchain_community.callbacks.manager import openai_callback_var
from langchain_core.callbacks import AsyncCallbackHandler
from langchain_core.callbacks.base import BaseCallbackHandler
from langchain_core.language_models.base import BaseLanguageModel
from langchain_openai.chat_models.azure import AzureChatOpenAI
from langchain_openai.chat_models.base import ChatOpenAI

from leaf_common.asyncio.asyncio_executor import AsyncioExecutor

from neuro_san.internals.interfaces.invocation_context import InvocationContext
from neuro_san.internals.journals.originating_journal import OriginatingJournal
from neuro_san.internals.messages.agent_message import AgentMessage
from neuro_san.internals.messages.origination import Origination
from neuro_san.internals.run_context.langchain.get_llm_token_callback import get_llm_token_callback
from neuro_san.internals.run_context.langchain.get_llm_token_callback import llm_token_callback_var
from neuro_san.internals.run_context.langchain.llm_token_callback_handler import LlmTokenCallbackHandler


# Keep a ContextVar for the origin info.  We do this because the
# langchain callbacks this stuff is based on also uses ContextVars
# and we want to be sure these are in sync.
# See: https://docs.python.org/3/library/contextvars.html
ORIGIN_INFO: ContextVar[str] = ContextVar('origin_info', default=None)


class LangChainTokenCounter:
    """
    Helps with per-llm means of counting tokens.
    Main entrypoint is count_tokens().

    Notes as to how each BaseLanguageModel/BaseChatModel should be configured
    are in get_callback_for_llm()
    """

    def __init__(self, llm: BaseLanguageModel,
                 invocation_context: InvocationContext,
                 journal: OriginatingJournal):
        """
        Constructor

        :param llm: The Llm to monitor for tokens
        :param invocation_context: The InvocationContext
        :param journal: The OriginatingJournal which through which this
                    will send token count AGENT messages
        """
        self.llm: BaseLanguageModel = llm
        self.invocation_context: InvocationContext = invocation_context
        self.journal: OriginatingJournal = journal
        self.debug: bool = False

    async def count_tokens(self, awaitable: Awaitable) -> Any:
        """
        Counts the tokens (if possible) from what happens inside the awaitable
        within a separate context.  If tokens are counted, they are added to
        the InvocationContext's request_reporting and sent over the message queue
        via the journal

        Recall awaitables are a full async method call with args.  That is, where you would expect to
                baz = await myinstance.foo(bar)
        you instead do
                baz = await token_counter.count_tokens(myinstance.foo(bar)).

        :param awaitable: The awaitable whose tokens we wish to count.
        :return: Whatever the awaitable would return
        """

        retval: Any = None

        # Take a time stamp so we measure another thing people care about - latency.
        start_time: float = time()

        # Attempt to count tokens/costs while invoking the agent.
        # The means by which this happens is on a per-LLM basis, so get the right hook
        # given the LLM we've got.
        callback: Union[AsyncCallbackHandler, BaseCallbackHandler] = None
        token_counter_context_manager = self.get_callback_for_llm(self.llm)

        if token_counter_context_manager is not None:

            # Record origin information in our own context var so we can associate
            # with the langchain callback context vars more easily.
            origin: List[Dict[str, Any]] = self.journal.get_origin()
            origin_str: str = Origination.get_full_name_from_origin(origin)
            ORIGIN_INFO.set(origin_str)

            old_callback: Union[AsyncCallbackHandler, BaseCallbackHandler] = None
            callback_var: ContextVar = self.get_context_var_for_llm(self.llm)
            if callback_var is not None:
                old_callback = callback_var.get()

            # Use the context manager to count tokens as per
            #   https://python.langchain.com/docs/how_to/llm_token_usage_tracking/#using-callbacks
            #
            # Caveats:
            # * In using this context manager approach, any tool that is called
            #   also has its token counts contributing to its callers for better or worse.
            # * As of 2/21/25, it seems that tool-calling agents (branch nodes) are not
            #   registering their tokens correctly. Not sure if this is a bug in langchain
            #   or there is something we are not doing in that scenario that we should be.
            with token_counter_context_manager() as callback:
                # Create a new context for different ContextVar values
                # and use the create_task() to run within that context.
                new_context: Context = copy_context()
                task: Task = new_context.run(self.create_task, awaitable)
                retval = await task

            if callback_var is not None:
                callback_var.set(old_callback)
        else:
            # No token counting was available for the LLM, but we still need to invoke.
            retval = await awaitable

        # Figure out how much time our agent took.
        end_time: float = time()
        time_taken_in_seconds: float = end_time - start_time

        await self.report(callback, time_taken_in_seconds)

        return retval

    def create_task(self, awaitable: Awaitable) -> Task:
        """
        Riffed from:
        https://stackoverflow.com/questions/78659844/async-version-of-context-run-for-context-vars-in-python-asyncio
        """
        executor: AsyncioExecutor = self.invocation_context.get_asyncio_executor()
        origin_str: str = ORIGIN_INFO.get()
        task: Task = executor.create_task(awaitable, origin_str)

        if self.debug:
            # Print to be sure we have a different callback object.
            oai_call = openai_callback_var.get()
            print(f"origin is {origin_str} callback var is {id(oai_call)}")

        return task

    async def report(self, callback: Union[AsyncCallbackHandler, BaseCallbackHandler], time_taken_in_seconds: float):
        """
        Report on the token accounting results of the callback

        :param callback: An AsyncCallbackHandler or BaseCallbackHandle instance that contains token counting information
        :param time_taken_in_seconds: The amount of time the awaitable took in count_tokens()
        """
        # Token counting results are collected in the callback, if there are any.
        # Different LLMs can count things in different ways, so normalize.
        token_dict: Dict[str, Any] = self.normalize_token_count(callback, time_taken_in_seconds)
        if token_dict is None or not bool(token_dict):
            return

        # Accumulate what we learned about tokens to request reporting.
        # For now we just overwrite the one key because we know
        # the last one out will be the front man, and as of 2/21/25 his stats
        # are cumulative.  At some point we might want a finer-grained breakdown
        # that perhaps contributes to a service/er-wide periodic token stats breakdown
        # of some kind.  For now, get something going.
        request_reporting: Dict[str, Any] = self.invocation_context.get_request_reporting()
        request_reporting["token_accounting"] = token_dict

        if self.journal is not None:
            # We actually have a token dictionary to report, so go there.
            agent_message = AgentMessage(structure=token_dict)
            await self.journal.write_message(agent_message)

    @staticmethod
    def get_callback_for_llm(llm: BaseLanguageModel) -> Any:
        """
        :param llm: A BaseLanguageModel returned from an LlmFactory.
        :return: A handle to a no-args function, that when called will
                open up a context manager for token counting.
                If not an OpenAI or Anthropic model, use context manager
                for LlmTokenCallbackHandler, which also gets token usage
                from "usage_metadata" but give "total_cost" = 0.
        """

        if isinstance(llm, (ChatOpenAI, AzureChatOpenAI)):
            # Notes:
            #   * ChatOpenAI needs to have stream_usage=True configured
            #     in order to get good token info back reliably.
            #   * AzureChatOpenAI needs to have model_kwargs.stream_options.include_usage=True
            #     configured in order to get good token info back reliably.
            return get_openai_callback

        # * ChatAnthropic needs to have stream_usage=True configured
        #   in order to get good token info back reliably.
        #   Per class docs this is on by default.

        # Open up a context manager for LlmTokenCallbackHandler, which also gets token usage
        # from "usage_metadata"
        # # Cost for Anthropic models can be determined as we import the lookup table
        # from https://python.langchain.com/api_reference/_modules/langchain_community/callbacks/
        # bedrock_anthropic_callback.html#BedrockAnthropicTokenUsageCallbackHandler
        # for non "claude" model "total_cost" = 0.0.
        return get_llm_token_callback

    @staticmethod
    def get_context_var_for_llm(llm: BaseLanguageModel) -> ContextVar:
        """
        :param llm: A BaseLanguageModel returned from an LlmFactory.
        :return: A ContextVar that corresponds to where token counting callback
                information is going.
                If not an OpenAI or Anthropic model, use llm_token_callback_var.
        """

        if isinstance(llm, (ChatOpenAI, AzureChatOpenAI)):
            return openai_callback_var

        # Collect tokens for models other than OpenAI
        return llm_token_callback_var

    @staticmethod
    def normalize_token_count(callback: Union[AsyncCallbackHandler, BaseCallbackHandler],
                              time_taken_in_seconds: float
                              ) -> Dict[str, Any]:
        """
        Normalizes the values in the token counting callback into a standard dictionary

        :param callback: An AsyncCallbackHandler or BaseCallbackHandler instance that contains
                            token counting information
        :param time_taken_in_seconds: The amount of time the awaitable took in count_tokens()
        """

        token_dict: Dict[str, Any] = {
            "time_taken_in_seconds": time_taken_in_seconds
        }
        if callback is None:
            return token_dict

        if isinstance(
            callback,
            (OpenAICallbackHandler, LlmTokenCallbackHandler)
        ):
            # So far these two instances share the same reporting structure
            token_dict = {
                "total_tokens": callback.total_tokens,
                "prompt_tokens": callback.prompt_tokens,
                "completion_tokens": callback.completion_tokens,
                "successful_requests": callback.successful_requests,
                "total_cost": callback.total_cost,
                "time_taken_in_seconds": time_taken_in_seconds,
                "caveats": [
                    "Each LLM Branch Node also includes accounting for each of its callees.",
                ]
            }

        return token_dict




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT
# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from neuro_san.api.grpc import concierge_pb2 as neuro__san_dot_api_dot_grpc_dot_concierge__pb2


class ConciergeServiceStub(object):
    """The service provides system-wide information about resources available
    in current instance of neuro-san.
    """

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.List = channel.unary_unary(
                '/dev.cognizant_ai.neuro_san.api.grpc.concierge.ConciergeService/List',
                request_serializer=neuro__san_dot_api_dot_grpc_dot_concierge__pb2.ConciergeRequest.SerializeToString,
                response_deserializer=neuro__san_dot_api_dot_grpc_dot_concierge__pb2.ConciergeResponse.FromString,
                )


class ConciergeServiceServicer(object):
    """The service provides system-wide information about resources available
    in current instance of neuro-san.
    """

    def List(self, request, context):
        """Called when a client needs the information about available agents.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_ConciergeServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'List': grpc.unary_unary_rpc_method_handler(
                    servicer.List,
                    request_deserializer=neuro__san_dot_api_dot_grpc_dot_concierge__pb2.ConciergeRequest.FromString,
                    response_serializer=neuro__san_dot_api_dot_grpc_dot_concierge__pb2.ConciergeResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'dev.cognizant_ai.neuro_san.api.grpc.concierge.ConciergeService', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class ConciergeService(object):
    """The service provides system-wide information about resources available
    in current instance of neuro-san.
    """

    @staticmethod
    def List(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/dev.cognizant_ai.neuro_san.api.grpc.concierge.ConciergeService/List',
            neuro__san_dot_api_dot_grpc_dot_concierge__pb2.ConciergeRequest.SerializeToString,
            neuro__san_dot_api_dot_grpc_dot_concierge__pb2.ConciergeResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)



# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT

import logging
import threading
from typing import Dict
from typing import List

from neuro_san.internals.interfaces.tool_factory_provider import ToolFactoryProvider
from neuro_san.internals.interfaces.agent_state_listener import AgentStateListener
from neuro_san.internals.run_context.interfaces.agent_tool_factory import AgentToolFactory
from neuro_san.internals.tool_factories.single_agent_tool_factory_provider import SingleAgentToolFactoryProvider


class ServiceToolFactoryProvider(ToolFactoryProvider):
    """
    Service-wide provider of agents tools factories.
    This class is a global singleton containing
    a table of currently active tool factories for each agent registered to the service.
    Note: a mapping from an agent to its tools factory is dynamic,
    as we can change agents definitions at service run-time.
    """

    instance = None

    def __init__(self):
        self.agents_table: Dict[str, AgentToolFactory] = {}
        self.logger = logging.getLogger(self.__class__.__name__)
        self.lock = threading.Lock()
        self.listeners: List[AgentStateListener] = []

    @classmethod
    def get_instance(cls):
        """
        Get a singleton instance of this class
        """
        if not ServiceToolFactoryProvider.instance:
            ServiceToolFactoryProvider.instance = ServiceToolFactoryProvider()
        return ServiceToolFactoryProvider.instance

    def add_listener(self, listener: AgentStateListener):
        """
        Add a state listener to be notified when status of service agents changes.
        """
        self.listeners.append(listener)

    def remove_listener(self, listener: AgentStateListener):
        """
        Remove a state listener from registered set.
        """
        if listener in self.listeners:
            self.listeners.remove(listener)

    def add_agent_tool_registry(self, agent_name: str, registry: AgentToolFactory):
        """
        This method is a single point of entry in the service
        where we register a new "agent name + AgentToolFactory" pair in the service scope
        or notify the service that for existing agent its AgentToolFactory has been modified.
        """
        is_new: bool = False
        with self.lock:
            is_new = self.agents_table.get(agent_name, None) is None
            self.agents_table[agent_name] = registry
        # Notify listeners about this state change:
        # do it outside of internal lock
        for listener in self.listeners:
            if is_new:
                listener.agent_added(agent_name)
                self.logger.info("ADDED tool registry for agent %s", agent_name)
            else:
                listener.agent_modified(agent_name)
                self.logger.info("REPLACED tool registry for agent %s", agent_name)

    def setup_tool_registries(self, registries: Dict[str, AgentToolFactory]):
        """
        Replace agents registries with "registries" collection.
        Previous state could be empty.
        """
        prev_agents: Dict[str, AgentToolFactory] = {}
        with self.lock:
            prev_agents = self.agents_table
            self.agents_table = {}
        # Notify listeners that previous set of agents is removed
        for agent_name, _ in prev_agents.items():
            for listener in self.listeners:
                listener.agent_removed(agent_name)
        # Add the new agent+registry pairs
        for agent_name, tool_registry in registries.items():
            self.add_agent_tool_registry(agent_name, tool_registry)

    def remove_agent_tool_registry(self, agent_name: str):
        """
        Remove agent name and its AgentToolFactory from service scope,
        so that agent becomes unavailable on our server.
        """
        with self.lock:
            self.agents_table.pop(agent_name, None)
        # Notify listeners about this state change:
        # do it outside of internal lock
        for listener in self.listeners:
            listener.agent_removed(agent_name)
        self.logger.info("REMOVED tool registry for agent %s", agent_name)

    def get_agent_tool_factory_provider(self, agent_name: str) -> ToolFactoryProvider:
        """
        Get agent tool factory provider for a specific agent
        :param agent_name: name of an agent
        """
        return SingleAgentToolFactoryProvider(agent_name, self.agents_table)

    def get_agent_names(self) -> List[str]:
        """
        Return static list of agent names.
        """
        with self.lock:
            # Create static snapshot of agents names collection
            return list(self.agents_table.keys())




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT
# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from neuro_san.api.grpc import agent_pb2 as neuro__san_dot_api_dot_grpc_dot_agent__pb2


class AgentServiceStub(object):
    """The service comprises all the exchanges to the backend in support of a single agent's
    services.  Routing is done by way of agent name on the grpc service hosting the agent,
    so as to keep info about which agents are hosted private (grpc gives the hand when a
    particular agent is unknown.
    """

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.Function = channel.unary_unary(
                '/dev.cognizant_ai.neuro_san.api.grpc.agent.AgentService/Function',
                request_serializer=neuro__san_dot_api_dot_grpc_dot_agent__pb2.FunctionRequest.SerializeToString,
                response_deserializer=neuro__san_dot_api_dot_grpc_dot_agent__pb2.FunctionResponse.FromString,
                )
        self.StreamingChat = channel.unary_stream(
                '/dev.cognizant_ai.neuro_san.api.grpc.agent.AgentService/StreamingChat',
                request_serializer=neuro__san_dot_api_dot_grpc_dot_agent__pb2.ChatRequest.SerializeToString,
                response_deserializer=neuro__san_dot_api_dot_grpc_dot_agent__pb2.ChatResponse.FromString,
                )
        self.Connectivity = channel.unary_unary(
                '/dev.cognizant_ai.neuro_san.api.grpc.agent.AgentService/Connectivity',
                request_serializer=neuro__san_dot_api_dot_grpc_dot_agent__pb2.ConnectivityRequest.SerializeToString,
                response_deserializer=neuro__san_dot_api_dot_grpc_dot_agent__pb2.ConnectivityResponse.FromString,
                )


class AgentServiceServicer(object):
    """The service comprises all the exchanges to the backend in support of a single agent's
    services.  Routing is done by way of agent name on the grpc service hosting the agent,
    so as to keep info about which agents are hosted private (grpc gives the hand when a
    particular agent is unknown.
    """

    def Function(self, request, context):
        """Called when a client needs the function description of an agent.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def StreamingChat(self, request, context):
        """Most important semantics of the streaming:
        1) The "answer" to a query of any agent network is in the *last* streamed
        AGENT_FRAMEWORK message.
        2) To RESTfully continue your conversation with the agent network:
        The very last AGENT_FRAMEWORK message before the stream closes will
        have its chat_context field filled in with a structure. You can
        copy this whole-cloth to the chat_context of your next StreamingChat
        request to continue the conversation.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Connectivity(self, request, context):
        """Called when a client needs the internal connectivity description of an agent.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_AgentServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'Function': grpc.unary_unary_rpc_method_handler(
                    servicer.Function,
                    request_deserializer=neuro__san_dot_api_dot_grpc_dot_agent__pb2.FunctionRequest.FromString,
                    response_serializer=neuro__san_dot_api_dot_grpc_dot_agent__pb2.FunctionResponse.SerializeToString,
            ),
            'StreamingChat': grpc.unary_stream_rpc_method_handler(
                    servicer.StreamingChat,
                    request_deserializer=neuro__san_dot_api_dot_grpc_dot_agent__pb2.ChatRequest.FromString,
                    response_serializer=neuro__san_dot_api_dot_grpc_dot_agent__pb2.ChatResponse.SerializeToString,
            ),
            'Connectivity': grpc.unary_unary_rpc_method_handler(
                    servicer.Connectivity,
                    request_deserializer=neuro__san_dot_api_dot_grpc_dot_agent__pb2.ConnectivityRequest.FromString,
                    response_serializer=neuro__san_dot_api_dot_grpc_dot_agent__pb2.ConnectivityResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'dev.cognizant_ai.neuro_san.api.grpc.agent.AgentService', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class AgentService(object):
    """The service comprises all the exchanges to the backend in support of a single agent's
    services.  Routing is done by way of agent name on the grpc service hosting the agent,
    so as to keep info about which agents are hosted private (grpc gives the hand when a
    particular agent is unknown.
    """

    @staticmethod
    def Function(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/dev.cognizant_ai.neuro_san.api.grpc.agent.AgentService/Function',
            neuro__san_dot_api_dot_grpc_dot_agent__pb2.FunctionRequest.SerializeToString,
            neuro__san_dot_api_dot_grpc_dot_agent__pb2.FunctionResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def StreamingChat(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(request, target, '/dev.cognizant_ai.neuro_san.api.grpc.agent.AgentService/StreamingChat',
            neuro__san_dot_api_dot_grpc_dot_agent__pb2.ChatRequest.SerializeToString,
            neuro__san_dot_api_dot_grpc_dot_agent__pb2.ChatResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def Connectivity(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/dev.cognizant_ai.neuro_san.api.grpc.agent.AgentService/Connectivity',
            neuro__san_dot_api_dot_grpc_dot_agent__pb2.ConnectivityRequest.SerializeToString,
            neuro__san_dot_api_dot_grpc_dot_agent__pb2.ConnectivityResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT
from typing import Any
from typing import Dict
from typing import Type

from neuro_san.internals.filters.maximal_message_filter import MaximalMessageFilter
from neuro_san.internals.filters.message_filter import MessageFilter
from neuro_san.internals.filters.minimal_message_filter import MinimalMessageFilter

TYPE_TO_MESSAGE_FILTER_CLASS: Dict[Any, Type[MessageFilter]] = {
    0:  MinimalMessageFilter,
    1:  MinimalMessageFilter,
    2:  MaximalMessageFilter,

    "UNKNOWN":  MinimalMessageFilter,
    "MINIMAL":  MinimalMessageFilter,
    "MAXIMAL":  MaximalMessageFilter,
}


class MessageFilterFactory:
    """
    Class for creating MessageFilters
    """

    @staticmethod
    def create_message_filter(chat_filter: Dict[str, Any]) -> MessageFilter:
        """
        :param chat_filter: The ChatFilter dictionary to process.
        :return: A MessageFilter that corresponds to the contents
        """

        # For now the default is MAXIMAL simply to emulate current behavior.
        # After the ChatFilter API gets released this will eventually change to "MINIMAL".
        default: str = "MINIMAL"
        chat_filter_type: Any = default

        # Get what was in the request
        if chat_filter is not None:
            chat_filter_type = chat_filter.get("chat_filter_type", chat_filter_type)

        # Change strings so they are suitable for lookup
        if isinstance(chat_filter_type, str):
            chat_filter_type = chat_filter_type.upper()

        chat_filter_class: MessageFilter = TYPE_TO_MESSAGE_FILTER_CLASS.get(chat_filter_type)
        if chat_filter_class is None:
            chat_filter_class = TYPE_TO_MESSAGE_FILTER_CLASS.get(default)

        # Instantiate the class and return
        message_filter: MessageFilter = chat_filter_class()
        return message_filter




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT
from typing import Any
from typing import Dict

from neuro_san.internals.run_context.interfaces.callable_tool import CallableTool
from neuro_san.internals.run_context.interfaces.run_context import RunContext


class AgentToolFactory:
    """
    Interface describing a factory that creates agent tools.
    Having this interface breaks some circular dependencies.
    """

    # pylint: disable=too-many-arguments,too-many-positional-arguments
    def create_agent_tool(self,
                          parent_run_context: RunContext,
                          parent_agent_spec: Dict[str, Any],
                          name: str,
                          sly_data: Dict[str, Any],
                          arguments: Dict[str, Any]) -> CallableTool:
        """
        Create an active node for an agent from its spec.

        :param parent_run_context: The RunContext of the agent calling this method
        :param parent_agent_spec: The spec of the agent calling this method.
        :param name: The name of the agent to get out of the registry
        :param sly_data: A mapping whose keys might be referenceable by agents, but whose
                 values should not appear in agent chat text. Can be an empty dictionary.
        :param arguments: A dictionary of arguments for the newly constructed agent
        :return: The CallableTool agent referred to by the name.
        """
        raise NotImplementedError

    def get_agent_tool_spec(self, name: str) -> Dict[str, Any]:
        """
        :param name: The name of the agent tool to get out of the registry
        :return: The dictionary representing the spec registered agent
        """
        raise NotImplementedError

    def get_config(self) -> Dict[str, Any]:
        """
        :return: The config dictionary given to the constructor.
        """
        raise NotImplementedError

    def get_agent_tool_path(self) -> str:
        """
        :return: The path under which tools for this registry should be looked for.
        """
        raise NotImplementedError

    def get_name_from_spec(self, agent_spec: Dict[str, Any]) -> str:
        """
        :param agent_spec: A single agent to register
        :return: The agent name as per the spec
        """
        raise NotImplementedError




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT
from typing import Any
from typing import Dict
from typing import List
from typing import Sequence
from typing import Union

import os
import json
import logging

from pathlib import Path
from pyparsing.exceptions import ParseException
from pyparsing.exceptions import ParseSyntaxException

from leaf_common.persistence.easy.easy_hocon_persistence import EasyHoconPersistence
from leaf_common.persistence.interface.restorer import Restorer

from neuro_san.internals.graph.persistence.agent_tool_registry_restorer import AgentToolRegistryRestorer
from neuro_san.internals.graph.registry.agent_tool_registry import AgentToolRegistry
from neuro_san.internals.utils.file_of_class import FileOfClass


class RegistryManifestRestorer(Restorer):
    """
    This interface provides a way to retrieve an object
    from some storage like a file, a database or S3.
    """

    def __init__(self, manifest_files: Union[str, List[str]] = None):
        """
        Constructor

        :param manifest_files: Either:
            * A single local name for the manifest file listing the agents to host.
            * A list of local names for multiple manifest files to host
            * None (the default) which gets a single manifest file from a known source.
        """
        self.manifest_files: List[str] = []

        if manifest_files is None:
            # We have no manifest list coming in, so check an env variable for a definition.
            manifest_file: str = os.environ.get("AGENT_MANIFEST_FILE")
            if manifest_file is None:
                # No env var, so fallback to whatis coded in this repo.
                file_of_class = FileOfClass(__file__, path_to_basis="../../../registries")
                manifest_file = file_of_class.get_file_in_basis("manifest.hocon")

            # Add what was found above
            self.manifest_files.append(manifest_file)
        elif isinstance(manifest_files, str):
            self.manifest_files.append(manifest_files)
        else:
            self.manifest_files = manifest_files

        self.logger = logging.getLogger(self.__class__.__name__)

    # pylint: disable=too-many-locals
    def restore_from_files(self, file_references: Sequence[str]) -> Dict[str, AgentToolRegistry]:
        """
        :param file_references: The sequence of file references to use when restoring.
        :return: a built map of agents tools registries
        """

        tool_registries: Dict[str, AgentToolRegistry] = {}

        # Loop through all the manifest files in the list to make a composite
        for manifest_file in file_references:

            one_manifest: Dict[str, Any] = {}
            if manifest_file.endswith(".hocon"):
                hocon = EasyHoconPersistence()
                try:
                    one_manifest = hocon.restore(file_reference=manifest_file)
                except (ParseException, ParseSyntaxException) as exception:
                    message: str = f"""
There was an error parsing the agent network manifest file "{manifest_file}".
See the accompanying ParseException (above) for clues as to what might be
syntactically incorrect in that file.
"""
                    raise ParseException(message) from exception
            else:
                try:
                    with open(manifest_file, "r", encoding="utf-8") as json_file:
                        one_manifest = json.load(json_file)
                except FileNotFoundError:
                    # Use the common verbiage below
                    one_manifest = None
                except json.decoder.JSONDecodeError as exception:
                    message: str = f"""
There was an error parsing the agent network manifest file "{manifest_file}".
See the accompanying JSONDecodeError exception (above) for clues as to what might be
syntactically incorrect in that file.
"""
                    raise ParseException(message) from exception

            if one_manifest is None:
                message = f"Could not find manifest file at path: {manifest_file}.\n" + """
Some common problems include:
* The file itself simply does not exist.
* Path is not an absolute path and you are invoking the server from a place
  where the path is not reachable.
* The path has a typo in it.

Double-check the value of the AGENT_MANIFEST_FILE env var and
your current working directory (pwd).
"""
                raise FileNotFoundError(message)

            for key, value in one_manifest.items():

                # Keys sometimes come with quotes.
                use_key: str = key.replace(r'"', "")

                if not bool(value):
                    continue

                file_of_class = FileOfClass(manifest_file)
                manifest_dir: str = file_of_class.get_basis()
                registry_restorer = AgentToolRegistryRestorer(manifest_dir)
                try:
                    tool_registry: AgentToolRegistry = registry_restorer.restore(file_reference=use_key)
                except FileNotFoundError as exc:
                    self.logger.error("Failed to restore registry item %s - %s", use_key, str(exc))
                    tool_registry = None
                if tool_registry is not None:
                    tool_name: str = Path(use_key).stem
                    tool_registries[tool_name] = tool_registry
                else:
                    self.logger.error("manifest registry %s not found in %s", use_key, manifest_file)

        return tool_registries

    # pylint: disable=too-many-locals
    def restore(self, file_reference: str = None) -> Dict[str, AgentToolRegistry]:
        """
        :param file_reference: The file reference to use when restoring.
                Default is None, implying the file reference is up to the
                implementation.
        :return: a built map of agents tools registries
        """
        if file_reference is not None:
            return self.restore_from_files([file_reference])

        tool_registries: Dict[str, AgentToolRegistry] =\
            self.restore_from_files(self.manifest_files)
        return tool_registries

    def get_manifest_files(self) -> List[str]:
        """
        Return current list of manifest files.
        """
        return self.manifest_files




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT

from collections.abc import Generator
from contextlib import contextmanager
from contextvars import ContextVar
from typing import Optional

from langchain_core.tracers.context import register_configure_hook

from neuro_san.internals.run_context.langchain.llm_token_callback_handler import LlmTokenCallbackHandler


llm_token_callback_var: ContextVar[Optional[LlmTokenCallbackHandler]] = (
        ContextVar("llm_token_callback", default=None)
    )
register_configure_hook(llm_token_callback_var, inheritable=True)


@contextmanager
def get_llm_token_callback() -> Generator[LlmTokenCallbackHandler, None, None]:
    """Get llm token callback.

    Get context manager for tracking usage metadata across chat model calls using
    "AIMessage.usage_metadata".

    This class is a modification of LangChain’s "UsageMetadataCallbackHandler":
    - https://python.langchain.com/api_reference/_modules/langchain_core/callbacks/usage.html
    #get_usage_metadata_callback
    """
    cb = LlmTokenCallbackHandler()
    llm_token_callback_var.set(cb)
    yield cb
    llm_token_callback_var.set(None)




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT
from typing import Any

from neuro_san.interfaces.async_agent_session import AsyncAgentSession


class AsyncAgentSessionFactory:
    """
    Creates asynchronous AsyncAgentSessions for external agents.
    """

    def create_session(self, agent_url: str, invocation_context: Any) -> AsyncAgentSession:
        """
        :param agent_url: A url string pointing to an external agent that came from
                    a tools list in an agent spec.
        :param invocation_context: The context policy container that pertains to the invocation
                    of the agent.

                    Note: At this interface level we are typing this as Any to avoid
                    an import cycle.  This will always be an InvocationContext.

        :return: An implementation of AsyncAgentSession through which
                 communications about external agents can be made.
        """
        raise NotImplementedError




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT
from __future__ import annotations

from typing import Any
from typing import Dict
from typing import Optional
from typing import Type

import logging
import traceback

from pydantic_core import ValidationError
from langchain_core.tools import BaseTool

from neuro_san.internals.run_context.interfaces.tool_caller import ToolCaller
from neuro_san.internals.run_context.langchain.langchain_run import LangChainRun
from neuro_san.internals.run_context.langchain.base_model_dictionary_converter \
    import BaseModelDictionaryConverter
from neuro_san.internals.run_context.langchain.pydantic_argument_dictionary_converter \
    import PydanticArgumentDictionaryConverter
from neuro_san.internals.run_context.utils.external_agent_parsing import ExternalAgentParsing


class LangChainOpenAIFunctionTool(BaseTool):
    """
    Class which represents an OpenAI Function as a langchain BaseTool.
    This is intended to partly be the reverse of format_tool_to_openai_function()

    This might seem antithetical as this just gets converted back
    to OpenAI function JSON in the langchain internals, but the idea
    is to have something at a higher level that plays well with
    the langchain Agent infrastructure.

    When this Tool's _arun() gets called, this allows the
    CallingTool (as ToolCaller) to set up other instances
    of LLM/Agent/AgentExecutor/RunContext to handle the tool call
    within that new context, but still within the async method chain.
    """

    # Declarations of member variables here satisfy Pydantic style,
    # which is a type validator that langchain is based on which
    # is able to use JSON schema definitions to validate fields.
    #
    # Also note that any of the function_json's parameters has to satisfy the
    # non-optional schema spec of this object. That is it needs:
    #       "name", "description" and "parameters" all as keys.

    # This group of member variables are required in OpenAI function definitions
    name: str
    description: str

    # All the other "member" variables assigned below are typed as
    # "Optional" so that we can have this class function as we like.

    parameters: Optional[Dict[str, Any]] = None

    # This guy is required to satisfy langchain internals.
    # See comment near assignment in from_function_json() below.
    args_schema: Optional[Type[BaseTool]] = None

    # The actual schema we want to report
    function_json: Optional[Dict[str, Any]] = None

    tool_caller: Optional[ToolCaller] = None

    @staticmethod
    def verify_function_json(function_json: Dict[str, Any]):
        """
        :param function_json: The function json description of the tool.
            Must have at least:
                * A name
                * A description
                * A parameters dictionary with at least one entry in its properties dict
        """
        message: str = ""
        name: str = function_json.get("name")
        if name is None:
            message = "function dictionary has no 'name' defined."
            raise ValueError(message)

        if function_json.get("description") is None:
            message = f"Function for {name} has no description.\n"

        parameters: Dict[str, Any] = function_json.get("parameters")
        if parameters:
            if parameters.get("type") is None:
                message = f"Function for {name} needs to have a parameters.type set to 'object'.\n"
            properties: Dict[str, Any] = parameters.get("properties")
            if properties is None:
                message = f"Function for {name} needs to have a properties dictionary as part of its parameters.\n"
            elif not isinstance(properties, Dict) or len(properties.keys()) == 0:
                message = f"Function for {name} needs to have at least one argument dictionary in its properties.\n"

        if len(message) > 0:
            message += """
This most often happens when calling an /external agent for the first time
and the hocon file for the agent network does not have a full function definition
specified for its front man for it to be called by another agent.
"""
            raise ValueError(message)

    @staticmethod
    def from_function_json(function_json: Dict[str, Any],
                           tool_caller: ToolCaller) \
            -> LangChainOpenAIFunctionTool:
        """
        Creates an instance of this tool from prepared OpenAI Function JSON.
        """
        # This first call populates the name, description and parameters
        # members of this BaseTool implementation so that it satisfies
        # the actual description of the function_json that we expect to
        # be passed in which we expect to conform to an OpenAI function.
        # definition.
        LangChainOpenAIFunctionTool.verify_function_json(function_json)
        try:
            tool = LangChainOpenAIFunctionTool(**function_json)
        except ValidationError as exception:
            message: str = f"""
Could not create tool to call extenal agent {function_json.get("name")}.
It's function_json is described thusly:
{function_json}
"""
            raise ValueError(message) from exception

        # Check for external tools.
        name: str = function_json.get("name")
        if ExternalAgentParsing.is_external_agent(name):
            tool.name = ExternalAgentParsing.get_safe_agent_name(name)

        # These next post-assignments satisfy the requirements of
        # convert_pydantic_to_openai_function() in that it wants to call
        # our schema() method to find out the function_json.
        use_function_json = function_json.get("parameters", function_json)
        tool.function_json = use_function_json

        # Langchain tools really prefer statically defined functions.
        # They are able to use some pretty fancy annotations and pydantic in order
        # to glean the function arguments and their types. We need more dynamism, though.
        #
        # The langchain_core BaseTool code for get_input_schema() expects there
        # to be an args_schema member which should be a pydantic BaseModel
        # for the objects that are passed as arguments.
        #
        # If "parameters" is not None, create a pydantic BaseModel of the OpenAI function spec for the agent
        # to satisfy that langchain need.  It's kind of a shame because this is just
        # going to get converted back to an OpenAI function again later on in langchain
        #  agent-land.
        if use_function_json != function_json:
            converter = BaseModelDictionaryConverter("parameters")
            tool.args_schema = converter.from_dict(use_function_json)

        tool.tool_caller = tool_caller

        return tool

    def _run(
        self,
        *args: Any,
        **kwargs: Any,
    ) -> Any:
        raise NotImplementedError

    async def _arun(
        self,
        *args: Any,
        **kwargs: Any,
    ) -> Any:
        """
        Comment from lanchain's BaseTool:
        Use the tool.

        Add run_manager: Optional[AsyncCallbackManagerForToolRun] = None
        to child implementations to enable tracing,

        Commentary:

        This method actually does the work when the tool is called
        via the AgentExecutor framework.
        It uses the tool_caller member as a basis to call the
        make_tool_function_calls() method. That guy
        is going to go through all the impending calls to Tools
        and do them all in a separate AgentExecutor context.
        Since this method is a langchain "a" method, all of these
        calls to another llm/agent instance happen asynchronously.
        ("a" is for asynchronous).

        :return: The result of the tool in the form of the BaseMessage
                 that tells us the "answer" from the tool.
        """
        run: LangChainRun = None
        try:
            # Use the CallingTool/tool_caller to invoke
            # its other tools within the context of other llm/agent/tool
            # instances.

            # When passing along the arguments, convert any pydantic-created objects
            # into dictionaries before passing them along, no matter how deep in an
            # object/dictionary hierarchy they may be.
            converter = PydanticArgumentDictionaryConverter()
            kwargs = converter.to_dict(kwargs)

            decision_run = LangChainRun("tool_base", [""], self.name, kwargs)
            run = await self.tool_caller.make_tool_function_calls(decision_run)

        # pylint: disable=broad-exception-caught
        except Exception as exception:
            # Report the exception, but return None as value from function.
            # This actually allows LLMs to recognize that something is wrong
            # and verbally report on that.
            logger = logging.getLogger(self.__class__.__name__)
            logger.error("Tool._arun() got Exception: %s", str(exception))
            logger.error(traceback.format_exc())
            run = None

        if run is None:
            return None

        # Assume the last message in the Run's chat_history holds the answer.
        # This eventually gets integrated into the CallingTool/tool_caller
        # in its submit_tool_outputs() call for its RunContext.
        the_answer = run.get_chat_history()[-1]
        return the_answer



# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT
from typing import Any

from unittest import TestCase

from neuro_san.test.interfaces.assert_forwarder import AssertForwarder


class UnitTestAssertForwarder(AssertForwarder):
    """
    AssertForwarder implemetation for python unittest.TestCase
    """

    def __init__(self, test_case: TestCase):
        """
        Constructor

        :param test_case: The python unitest.TestCase instance on which to base any asserts.
        """
        self.test_case: TestCase = test_case

    def assertEqual(self, first: Any, second: Any, msg: str = None):
        """
        Assert that the first is equal to the second

        :param first: First comparison element
        :param second: Second comparison element
        :param msg: optional string message
        """
        self.test_case.assertEqual(first, second, msg=msg)

    def assertNotEqual(self, first: Any, second: Any, msg: str = None):
        """
        Assert that the first is not equal to the second

        :param first: First comparison element
        :param second: Second comparison element
        :param msg: optional string message
        """
        self.test_case.assertNotEqual(first, second, msg=msg)

    def assertTrue(self, expr: Any, msg: str = None):
        """
        Assert that the expression is true

        :param expr: Expression to test
        :param msg: optional string message
        """
        self.test_case.assertTrue(expr, msg=msg)

    def assertFalse(self, expr: Any, msg: str = None):
        """
        Assert that the expression is false

        :param expr: Expression to test
        :param msg: optional string message
        """
        self.test_case.assertFalse(expr, msg=msg)

    def assertIs(self, first: Any, second: Any, msg: str = None):
        """
        Assert that the first and second are the same object

        :param first: First comparison element
        :param second: Second comparison element
        :param msg: optional string message
        """
        self.test_case.assertIs(first, second, msg=msg)

    def assertIsNot(self, first: Any, second: Any, msg: str = None):
        """
        Assert that the first and second are not the same object

        :param first: First comparison element
        :param second: Second comparison element
        :param msg: optional string message
        """
        self.test_case.assertIsNot(first, second, msg=msg)

    def assertIsNone(self, expr: Any, msg: str = None):
        """
        Assert that the expression is None

        :param expr: Expression to test
        :param msg: optional string message
        """
        self.test_case.assertIsNone(expr, msg=msg)

    def assertIsNotNone(self, expr: Any, msg: str = None):
        """
        Assert that the expression is not None

        :param expr: Expression to test
        :param msg: optional string message
        """
        self.test_case.assertIsNotNone(expr, msg=msg)

    def assertIn(self, member: Any, container: Any, msg: str = None):
        """
        Assert that the member is in the container

        :param member: Member comparison element
        :param container: Container comparison element
        :param msg: optional string message
        """
        self.test_case.assertIn(member, container, msg=msg)

    def assertNotIn(self, member: Any, container: Any, msg: str = None):
        """
        Assert that the member is not in the container

        :param member: Member comparison element
        :param container: Container comparison element
        :param msg: optional string message
        """
        self.test_case.assertNotIn(member, container, msg=msg)

    # pylint: disable=invalid-name
    def assertIsInstance(self, obj: Any, cls: Any, msg: str = None):
        """
        Assert that the obj is an instance of the cls

        :param obj: object instance comparison element
        :param cls: Class comparison element
        :param msg: optional string message
        """
        self.test_case.assertIsInstance(obj, cls, msg=msg)

    # pylint: disable=invalid-name
    def assertNotIsInstance(self, obj: Any, cls: Any, msg: str = None):
        """
        Assert that the obj is not an instance of the cls

        :param obj: object instance comparison element
        :param cls: Class comparison element
        :param msg: optional string message
        """
        self.test_case.assertNotIsInstance(obj, cls, msg=msg)

    # pylint: disable=invalid-name
    def assertGreater(self, first: Any, second: Any, msg: str = None):
        """
        Assert that the first is greater than the second.

        :param first: First comparison element
        :param second: Second comparison element
        :param msg: optional string message
        """
        self.test_case.assertGreater(first, second, msg=msg)

    # pylint: disable=invalid-name
    def assertGreaterEqual(self, first: Any, second: Any, msg: str = None):
        """
        Assert that the first is greater than or equal to the second.

        :param first: First comparison element
        :param second: Second comparison element
        :param msg: optional string message
        """
        self.test_case.assertGreaterEqual(first, second, msg=msg)

    # pylint: disable=invalid-name
    def assertLess(self, first: Any, second: Any, msg: str = None):
        """
        Assert that the first is less than the second.

        :param first: First comparison element
        :param second: Second comparison element
        :param msg: optional string message
        """
        self.test_case.assertLess(first, second, msg=msg)

    # pylint: disable=invalid-name
    def assertLessEqual(self, first: Any, second: Any, msg: str = None):
        """
        Assert that the first is less than or equal to the second.

        :param first: First comparison element
        :param second: Second comparison element
        :param msg: optional string message
        """
        self.test_case.assertLessEqual(first, second, msg=msg)

    # pylint: disable=invalid-name
    def assertGist(self, gist: bool, acceptance_criteria: str, text_sample: str, msg: str = None):
        """
        Assert that the gist is true

        :param gist: Pass/Fail value of the gist expected to be True
        :param acceptance_criteria: The value to verify against
        :param text_sample: The value appearing in the test sample
        :param msg: optional string message
        """
        if msg is None:
            msg = f"""
text_sample unexpectedly did not match acceptance criteria.

text_sample:
{text_sample}

acceptance_criteria:
{acceptance_criteria}
"""
        self.test_case.assertTrue(gist, msg=msg)

    # pylint: disable=invalid-name
    def assertNotGist(self, gist: bool, acceptance_criteria: str, text_sample: str, msg: str = None):
        """
        Assert that the gist is true

        :param gist: Pass/Fail value of the gist expected to be False
        :param acceptance_criteria: The value to verify against
        :param text_sample: The value appearing in the test sample
        :param msg: optional string message
        """
        if msg is None:
            msg = f"""
text_sample unexpectedly did match acceptance criteria.

text_sample:
{text_sample}

acceptance_criteria:
{acceptance_criteria}
"""
        self.test_case.assertFalse(gist, msg=msg)




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT
from typing import Any
from typing import Dict
from typing import List

from langchain_core.messages.base import BaseMessage


class Journal:
    """
    An interface for journaling chat messages.
    """

    async def write_message(self, message: BaseMessage, origin: List[Dict[str, Any]]):
        """
        Writes a BaseMessage entry into the journal
        :param message: The BaseMessage instance to write to the journal
        :param origin: A List of origin dictionaries indicating the origin of the run.
                The origin can be considered a path to the original call to the front-man.
                Origin dictionaries themselves each have the following keys:
                    "tool"                  The string name of the tool in the spec
                    "instantiation_index"   An integer indicating which incarnation
                                            of the tool is being dealt with.
        """
        raise NotImplementedError




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT
"""
See class comment for details
"""
from typing import Any, Dict

from neuro_san.http_sidecar.handlers.base_request_handler import BaseRequestHandler
from neuro_san.interfaces.async_agent_session import AsyncAgentSession


class ConnectivityHandler(BaseRequestHandler):
    """
    Handler class for neuro-san "connectivity" API call.
    """

    async def get(self, agent_name: str):
        """
        Implementation of GET request handler for "connectivity" API call.
        """
        metadata: Dict[str, Any] = self.get_metadata()
        update_done: bool = await self.update_agents(metadata)
        if not update_done:
            return

        if not self.agent_policy.allow(agent_name):
            self.set_status(404)
            self.logger.error({}, "error: Invalid request path %s", self.request.path)
            self.do_finish()
            return

        self.logger.info(metadata, "Start GET %s/connectivity", agent_name)
        try:
            data: Dict[str, Any] = {}
            grpc_session: AsyncAgentSession = self.get_agent_grpc_session(metadata, agent_name)
            result_dict: Dict[str, Any] = await grpc_session.connectivity(data)

            # Return gRPC response to the HTTP client
            self.set_header("Content-Type", "application/json")
            self.write(result_dict)

        except Exception as exc:  # pylint: disable=broad-exception-caught
            self.process_exception(exc)
        finally:
            self.do_finish()
            self.logger.info(metadata, "Finish GET %s/connectivity request", agent_name)




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT
from typing import Any
from typing import Dict

from asyncio import Future

from leaf_common.asyncio.asyncio_executor import AsyncioExecutor
from leaf_server_common.logging.logging_setup import setup_extra_logging_fields

from neuro_san.internals.chat.async_collating_queue import AsyncCollatingQueue
from neuro_san.internals.interfaces.async_agent_session_factory import AsyncAgentSessionFactory
from neuro_san.internals.interfaces.context_type_toolbox_factory import ContextTypeToolboxFactory
from neuro_san.internals.interfaces.context_type_llm_factory import ContextTypeLlmFactory
from neuro_san.internals.interfaces.invocation_context import InvocationContext
from neuro_san.internals.journals.message_journal import MessageJournal
from neuro_san.internals.journals.journal import Journal
from neuro_san.internals.messages.origination import Origination


# pylint: disable=too-many-instance-attributes
class SessionInvocationContext(InvocationContext):
    """
    Implementation of InvocationContext which encapsulates specific policy classes that pertain to
    a single invocation of an AgentSession, whether by way of a
    service call or library call.
    """

    def __init__(self, async_session_factory: AsyncAgentSessionFactory,
                 llm_factory: ContextTypeLlmFactory,
                 toolbox_factory: ContextTypeToolboxFactory = None,
                 metadata: Dict[str, str] = None):
        """
        Constructor

        :param async_session_factory: The AsyncAgentSessionFactory to use
                        when connecting with external agents.
        :param llm_factory: The ContextTypeLlmFactory instance
        :param metadata: A grpc metadata of key/value pairs to be inserted into
                         the header. Default is None. Preferred format is a
                         dictionary of string keys to string values.
        """

        self.async_session_factory: AsyncAgentSessionFactory = async_session_factory
        # Get an async executor to run all tasks for this session instance:
        self.asyncio_executor: AsyncioExecutor = AsyncioExecutor()
        self.origination: Origination = Origination()
        self.queue: AsyncCollatingQueue = AsyncCollatingQueue()
        self.journal: Journal = MessageJournal(self.queue)
        self.metadata: Dict[str, str] = metadata
        self.request_reporting: Dict[str, Any] = {}
        self.llm_factory: ContextTypeLlmFactory = llm_factory
        self.toolbox_factory: ContextTypeToolboxFactory = toolbox_factory

    def start(self):
        """
        Starts the active components of this invocation context.
        Do this separately from constructor for more control.
        Currently, we only start internal AsyncioExecutor.
        """
        self.asyncio_executor.start()

        # Set up logging fields within the thread, so we have consistent logging from async calls.
        # Ignore the future that is returned - we trust it will get done.
        _: Future = self.asyncio_executor.submit(
            "logging_setup", setup_extra_logging_fields,
            metadata_dict=self.metadata)

    def get_async_session_factory(self) -> AsyncAgentSessionFactory:
        """
        :return: The AsyncAgentSessionFactory associated with the invocation
        """
        return self.async_session_factory

    def get_asyncio_executor(self) -> AsyncioExecutor:
        """
        :return: The AsyncioExecutor associated with the invocation
        """
        return self.asyncio_executor

    def get_origination(self) -> Origination:
        """
        :return: The Origination instance carrying state about tool instantation
                during the course of the AgentSession.
        """
        return self.origination

    def get_journal(self) -> Journal:
        """
        :return: The Journal instance that allows message reporting
                during the course of the AgentSession.
        """
        return self.journal

    def get_queue(self) -> AsyncCollatingQueue:
        """
        :return: The AsyncCollatingQueue instance via which messages are streamed to the
                AgentSession mechanics
        """
        return self.queue

    def get_metadata(self) -> Dict[str, str]:
        """
        :return: The metadata to pass along with any request
        """
        return self.metadata

    def close(self):
        """
        Release resources owned by this context
        """
        if self.asyncio_executor is not None:
            self.asyncio_executor.shutdown()
            self.asyncio_executor = None

    def get_request_reporting(self) -> Dict[str, Any]:
        """
        :return: The request reporting dictionary
        """
        return self.request_reporting

    def get_llm_factory(self) -> ContextTypeLlmFactory:
        """
        :return: The ContextTypeLlmFactory instance for the session
        """
        return self.llm_factory

    def get_toolbox_factory(self) -> ContextTypeToolboxFactory:
        """
        :return: The ContextTypeToolboxFactory instance for the session
        """
        return self.toolbox_factory

    def reset(self):
        """
        Resets the instance for a subsequent use for another exchange with the agent network.
        """
        # Origination needs to be reset so that origin information can match up
        # with what is in the chat_context. If we do not reset this, then library calls
        # to DirectAgentSession do not properly carry forward any memory of the conversation
        # in subsequent interactions with the same network.
        self.origination.reset()



# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT
from typing import Any
from typing import Dict
from typing import List

from neuro_san.test.interfaces.null_assert_forwarder import NullAssertForwarder


class AssessorAssertForwarder(NullAssertForwarder):
    """
    AssertForwarder implemetation for the agent Assessor.

    This guy generally does not care about the correctness of the asserts themselves,
    but serves to collect failure data for the Assessor.
    """

    def __init__(self):
        """
        Constructor
        """
        self.num_total: int = 0
        self.fail: List[Dict[str, Any]] = []

    def get_num_total(self) -> int:
        """
        :return: The total number of test cases seen
        """
        return self.num_total

    def get_fail_dicts(self) -> List[Dict[str, Any]]:
        """
        :return: The failure dictionaries for further analysis
        """
        return self.fail

    # pylint: disable=invalid-name
    def assertGist(self, gist: bool, acceptance_criteria: str, text_sample: str, msg: str = None):
        """
        Assert that the gist is true

        :param gist: Pass/Fail value of the gist expected to be True
        :param acceptance_criteria: The value to verify against
        :param text_sample: The value appearing in the test sample
        :param msg: optional string message
        """
        self.handle_assert(gist, acceptance_criteria, text_sample, True)

    # pylint: disable=invalid-name
    def assertNotGist(self, gist: bool, acceptance_criteria: str, text_sample: str, msg: str = None):
        """
        Assert that the gist is true

        :param gist: Pass/Fail value of the gist expected to be False
        :param acceptance_criteria: The value to verify against
        :param text_sample: The value appearing in the test sample
        :param msg: optional string message
        """
        self.handle_assert(gist, acceptance_criteria, text_sample, False)

    def handle_assert(self, is_passing: bool,
                      acceptance_criteria: str,
                      text_sample: str,
                      sense: bool):
        """
        Handle the assert.
        :param is_passing: Boolean as to whether or not the test is passing.
        :param acceptance_criteria: The value to verify against
        :param text_sample: The value appearing in the test sample
        :param sense: Whether the test was supposed to be true or false.
        """
        self.num_total += 1
        if is_passing == sense:
            return

        components: Dict[str, Any] = {
            "acceptance_criteria": acceptance_criteria,
            "text_sample": text_sample,
            "sense": sense
        }
        self.fail.append(components)




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT
"""
See class comment for details
"""
from typing import Dict
from typing import List

import multiprocessing
import os

from argparse import ArgumentParser

from leaf_server_common.server.server_loop_callbacks import ServerLoopCallbacks

from neuro_san.interfaces.agent_session import AgentSession
from neuro_san.internals.graph.persistence.registry_manifest_restorer import RegistryManifestRestorer
from neuro_san.internals.graph.registry.agent_tool_registry import AgentToolRegistry
from neuro_san.service.agent_server import AgentServer
from neuro_san.service.agent_server import DEFAULT_SERVER_NAME
from neuro_san.service.agent_server import DEFAULT_SERVER_NAME_FOR_LOGS
from neuro_san.service.agent_server import DEFAULT_MAX_CONCURRENT_REQUESTS
from neuro_san.service.agent_server import DEFAULT_REQUEST_LIMIT
from neuro_san.service.agent_server import DEFAULT_FORWARDED_REQUEST_METADATA
from neuro_san.service.grpc_agent_service import GrpcAgentService
from neuro_san.http_sidecar.http_sidecar import HttpSidecar
from neuro_san.internals.utils.file_of_class import FileOfClass
from neuro_san.service.registries_watcher.periodic_updater.manifest_periodic_updater import ManifestPeriodicUpdater


# pylint: disable=too-many-instance-attributes
class AgentMainLoop(ServerLoopCallbacks):
    """
    This class handles the service main loop.
    """

    def __init__(self):
        """
        Constructor
        """
        self.port: int = 0
        self.http_port: int = 0

        self.tool_registries: Dict[str, AgentToolRegistry] = {}

        self.server_name: str = DEFAULT_SERVER_NAME
        self.server_name_for_logs: str = DEFAULT_SERVER_NAME_FOR_LOGS
        self.max_concurrent_requests: int = DEFAULT_MAX_CONCURRENT_REQUESTS
        self.request_limit: int = DEFAULT_REQUEST_LIMIT
        self.forwarded_request_metadata: int = DEFAULT_FORWARDED_REQUEST_METADATA
        self.service_openapi_spec_file: str = self._get_default_openapi_spec_path()
        self.manifest_update_period_seconds: int = 0
        self.server: AgentServer = None
        self.manifest_files: List[str] = []

    def parse_args(self):
        """
        Parse command-line arguments into member variables
        """
        # Set up the CLI parser
        arg_parser = ArgumentParser()

        arg_parser.add_argument("--port", type=int,
                                default=int(os.environ.get("AGENT_PORT", AgentSession.DEFAULT_PORT)),
                                help="Port number for the grpc service")
        arg_parser.add_argument("--http_port", type=int,
                                default=int(os.environ.get("AGENT_HTTP_PORT", AgentSession.DEFAULT_HTTP_PORT)),
                                help="Port number for http service endpoint")
        arg_parser.add_argument("--server_name", type=str,
                                default=str(os.environ.get("AGENT_SERVER_NAME", self.server_name)),
                                help="Name of the service for health reporting purposes.")
        arg_parser.add_argument("--server_name_for_logs", type=str,
                                default=str(os.environ.get("AGENT_SERVER_NAME_FOR_LOGS", self.server_name_for_logs)),
                                help="Name of the service as seen in logs")
        arg_parser.add_argument("--max_concurrent_requests", type=int,
                                default=int(os.environ.get("AGENT_MAX_CONCURRENT_REQUESTS",
                                                           self.max_concurrent_requests)),
                                help="Maximum number of requests that can be served at the same time")
        arg_parser.add_argument("--request_limit", type=int,
                                default=int(os.environ.get("AGENT_REQUEST_LIMIT", self.request_limit)),
                                help="Number of requests served before the server shuts down in an orderly fashion")
        arg_parser.add_argument("--forwarded_request_metadata", type=str,
                                default=os.environ.get("AGENT_FORWARDED_REQUEST_METADATA",
                                                       self.forwarded_request_metadata),
                                help="Space-delimited list of http request metadata keys to forward "
                                     "to logs/other requests")
        arg_parser.add_argument("--openapi_service_spec_path", type=str,
                                default=os.environ.get("AGENT_OPENAPI_SPEC",
                                                       self.service_openapi_spec_file),
                                help="File path to OpenAPI service specification document.")
        arg_parser.add_argument("--manifest_update_period_seconds", type=int,
                                default=int(os.environ.get("AGENT_MANIFEST_UPDATE_PERIOD_SECONDS", "0")),
                                help="Periodic run-time update period for manifest in seconds."
                                     " Value <= 0 disables updates.")

        # Actually parse the args into class variables

        # Incorrectly flagged as Path Traversal 3, 7
        # See destination below ~ line 139, 154 for explanation.
        args = arg_parser.parse_args()
        self.port = args.port
        self.http_port = args.http_port
        self.server_name = args.server_name
        self.server_name_for_logs = args.server_name_for_logs
        self.max_concurrent_requests = args.max_concurrent_requests
        self.request_limit = args.request_limit
        self.forwarded_request_metadata = args.forwarded_request_metadata
        self.service_openapi_spec_file = args.openapi_service_spec_path
        self.manifest_update_period_seconds = args.manifest_update_period_seconds

        manifest_restorer = RegistryManifestRestorer()
        manifest_tool_registries: Dict[str, AgentToolRegistry] = manifest_restorer.restore()
        self.manifest_files = manifest_restorer.get_manifest_files()

        self.tool_registries = manifest_tool_registries

    def _get_default_openapi_spec_path(self) -> str:
        """
        Return a file path to default location of OpenAPI specification file
        for neuro-san service.
        """
        file_of_class = FileOfClass(__file__, path_to_basis="../api/grpc")
        return file_of_class.get_file_in_basis("agent_service.json")

    def main_loop(self):
        """
        Command line entry point
        """
        self.parse_args()

        self.server = AgentServer(self.port,
                                  server_loop_callbacks=self,
                                  tool_registries=self.tool_registries,
                                  server_name=self.server_name,
                                  server_name_for_logs=self.server_name_for_logs,
                                  max_concurrent_requests=self.max_concurrent_requests,
                                  request_limit=self.request_limit,
                                  forwarded_request_metadata=self.forwarded_request_metadata)

        if self.manifest_update_period_seconds > 0:
            manifest_file: str = self.manifest_files[0]
            updater: ManifestPeriodicUpdater =\
                ManifestPeriodicUpdater(manifest_file, self.manifest_update_period_seconds)
            updater.start()

        # Start HTTP server side-car:
        http_sidecar = HttpSidecar(
            self.port,
            self.http_port,
            self.tool_registries,
            self.service_openapi_spec_file,
            forwarded_request_metadata=self.forwarded_request_metadata)
        http_server_process = multiprocessing.Process(target=http_sidecar)
        http_server_process.start()

        try:
            self.server.serve()
        finally:
            http_server_process.terminate()
            http_server_process.join()

    def loop_callback(self) -> bool:
        """
        Periodically called by the main server loop of ServerLifetime.
        """
        # Report back on service activity so the ServerLifetime that calls
        # this method can properly yield/sleep depending on how many requests
        # are in motion.
        agent_services: List[GrpcAgentService] = self.server.get_services()
        for agent_service in agent_services:
            if agent_service.get_request_count() > 0:
                return True

        return False


if __name__ == '__main__':
    AgentMainLoop().main_loop()




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT
# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: neuro_san/api/grpc/concierge.proto
# Protobuf Python Version: 4.25.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.api import annotations_pb2 as google_dot_api_dot_annotations__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\"neuro_san/api/grpc/concierge.proto\x12-dev.cognizant_ai.neuro_san.api.grpc.concierge\x1a\x1cgoogle/api/annotations.proto\"\x12\n\x10\x43onciergeRequest\"M\n\tAgentInfo\x12\x1e\n\nagent_name\x18\x01 \x01(\tR\nagent_name\x12 \n\x0b\x64\x65scription\x18\x02 \x01(\tR\x0b\x64\x65scription\"e\n\x11\x43onciergeResponse\x12P\n\x06\x61gents\x18\x01 \x03(\x0b\x32\x38.dev.cognizant_ai.neuro_san.api.grpc.concierge.AgentInfoR\x06\x61gents2\xb4\x01\n\x10\x43onciergeService\x12\x9f\x01\n\x04List\x12?.dev.cognizant_ai.neuro_san.api.grpc.concierge.ConciergeRequest\x1a@.dev.cognizant_ai.neuro_san.api.grpc.concierge.ConciergeResponse\"\x14\x82\xd3\xe4\x93\x02\x0e\x12\x0c/api/v1/listBfZdgithub.com/leaf-ai/neuro_san/internal/gen/dev.cognizant_ai/neuro_san/api/grpc/concierge/v1;conciergeb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'neuro_san.api.grpc.concierge_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  _globals['DESCRIPTOR']._options = None
  _globals['DESCRIPTOR']._serialized_options = b'Zdgithub.com/leaf-ai/neuro_san/internal/gen/dev.cognizant_ai/neuro_san/api/grpc/concierge/v1;concierge'
  _globals['_CONCIERGESERVICE'].methods_by_name['List']._options = None
  _globals['_CONCIERGESERVICE'].methods_by_name['List']._serialized_options = b'\202\323\344\223\002\016\022\014/api/v1/list'
  _globals['_CONCIERGEREQUEST']._serialized_start=115
  _globals['_CONCIERGEREQUEST']._serialized_end=133
  _globals['_AGENTINFO']._serialized_start=135
  _globals['_AGENTINFO']._serialized_end=212
  _globals['_CONCIERGERESPONSE']._serialized_start=214
  _globals['_CONCIERGERESPONSE']._serialized_end=315
  _globals['_CONCIERGESERVICE']._serialized_start=318
  _globals['_CONCIERGESERVICE']._serialized_end=498
# @@protoc_insertion_point(module_scope)
